//
// to test FIRST_k_{alpha} algorithm of Aho and Ullman.
// Created on: 2012/09/02
//

#include "y.h"
#include "lane_tracing.h"

#define DEBUG_FK 0 // 1, 2, 4 for more details. 0 - close all.


////////////////////////////////////
// Set for FIRST_K. START.
// Create the set as a linked list for now.
////////////////////////////////////

// // Declarations here are moved to lane_tracing.h.
// struct FK_SET_E {
//   SymbolList elem;
//   int length;
//   struct FK_SET_E * next;
// };
// typedef struct FK_SET_E FK_SET_Elem;
// 
// typedef struct {
//   FK_SET_Elem * head;
//   FK_SET_Elem * tail;
//   int count;
// } FK_SET;


void FK_SET_dump(FK_SET * s); // Defined at the end of this file.

static FK_SET * FK_SET_create() {
  FK_SET * s;
  HYY_NEW(s, FK_SET, 1);
  s->head = s->tail = NULL;
  s->count = 0;
  return s;
}

static void FK_SET_Elem_dump(FK_SET_Elem * e) {
  SymbolList s = e->elem;
  printf("[ ");
  for (; s != NULL; s = s->next) printf("%s ", s->snode->symbol);
  printf("]");
}

static void FK_SET_dump2(FK_SET * s) {
  FK_SET_Elem * e;
  int i = 0;
  if (s == NULL) return;

  e = s->head;
  printf("{");
  for (; e != NULL; e = e->next) {
    //printf("  %2d:", i++);
    FK_SET_Elem_dump(e);
  }
  //printf("} (set elem count: %d)", s->count);
  printf("}");
}

static FK_SET_Elem * FK_SET_Elem_create(SymbolNode * n) {
  FK_SET_Elem * e;
  HYY_NEW(e, FK_SET_Elem, 1);
  e->elem = n; // createSymbolNode(n);
  e->length = 1;
  e->next = NULL;
  return e;
}

static void FK_SET_Elem_destroy(FK_SET_Elem ** e) {
  if (*e == NULL) return;
  freeSymbolNodeList((*e)->elem);
  free(*e);
  *e = NULL;
}

static FK_SET_Elem * FK_SET_Elem_clone(FK_SET_Elem * e) {
  FK_SET_Elem * f = FK_SET_Elem_create( cloneSymbolList(e->elem) );
  f->length = e->length;
  f->next = e->next;
  return f;
}

//
// return 1 if e1 > e2, 
//       -1 if e1 < e2,
//        0 if e1 == e2.
//
static int FK_SET_Elem_compare(FK_SET_Elem * e1, FK_SET_Elem * e2) {
  SymbolNode * s1, * s2;
  int v;

  if (e1 == NULL && e2 == NULL) return 0;
  else if (e1 == NULL) return -1;
  else if (e2 == NULL) return 1;
  // else.

  v = e1->length - e2->length;
  if (v != 0) return v;

  for (s1 = e1->elem, s2 = e2->elem; s1 != NULL && s2 != NULL; 
       s1 = s1->next, s2 = s2->next) {
    if (s1->snode != s2->snode) {
      v = strcmp(s1->snode->symbol, s2->snode->symbol);
      if (0 != v) return v;
    }
  }

  if (s1 == NULL && s2 == NULL) return 0;
  else if (s1 != NULL) return 1;
  return -1;
}

static int FK_SET_compare(FK_SET * s1, FK_SET * s2) {
  FK_SET_Elem * e1, * e2;
  int v;

  for (e1 = s1->head, e2 = s2->head; e1 != NULL && e2 != NULL; 
       e1 = e1->next, e2 = e2->next) {
    v = FK_SET_Elem_compare(e1, e2);
    if (v != 0) return v;
  }

  if (e1 == NULL && e2 == NULL) return 0;
  else if (e1 != NULL) return 1;
  return -1;
}

//
// if e does not exist in s, insert it.
// Note: e will be used, but not a copy of e.
//
static FK_SET * FK_SET_insert(FK_SET * s, FK_SET_Elem * e) {
  int v;
  if (s == NULL) return NULL;

  if (s->count == 0) {
    s->head = s->tail = e; //FK_SET_Elem_create();
    e->next = NULL;
  } else {
    // get through list, insert if not exists yet.
    FK_SET_Elem * i, * i_prev;
    for (i_prev = NULL, i = s->head; i != NULL; i_prev = i, i = i->next) {
      v = FK_SET_Elem_compare(e, i);
      if (v == 0) return s; // e already in s. do nothing and return.
      if (v > 0) continue;
      else { // v < 0, insert e before i.
        if (i_prev == NULL) {
          e->next = s->head;
          s->head = e;
        } else {
          e->next = i;
          i_prev->next = e;
        }
        break;
      } // end of else
    } // end of for loop.
    if (i == NULL) { // insert e as the new tail.
      s->tail->next = e;
      s->tail = e;
      e->next = NULL;
    }
  }

  s->count ++;
  //printf("insert s->count=%d\n", s->count);
  return s;
}

static void FK_SET_destroy(FK_SET ** s) {
  FK_SET_Elem * e, * e_prev;
  if (*s == NULL) return;

  for (e_prev = NULL, e = (*s)->head; e != NULL; e_prev = e, e = e->next) {
    FK_SET_Elem_destroy(& e_prev);
  }
  FK_SET_Elem_destroy(& e_prev);
  
  free(*s);
  *s = NULL;
}

static FK_SET * FK_SET_clone(FK_SET * s) {
  FK_SET * t;
  FK_SET_Elem * e;
  if (s == NULL) return s;

  t = FK_SET_create();
  for (e = s->head; e != NULL; e = e->next) {
    FK_SET_insert(t, FK_SET_Elem_clone(e));
  }

  return t;
}

// Combine set s2 to set s1.
static FK_SET * FK_SET_union(FK_SET * s1, FK_SET * s2) {
  FK_SET_Elem * e;
  for (e = s2->head; e != NULL; e = e->next) {
    s1 = FK_SET_insert( s1, FK_SET_Elem_clone(e) );
  }
  return s1;
}

// Combine sets s1 and s2 to a NEW set.
static FK_SET * FK_SET_union_NEW(FK_SET * s1, FK_SET * s2) {
  FK_SET * s = FK_SET_create();
  FK_SET_Elem * e;
  for (e = s1->head; e != NULL; e = e->next) {
    s = FK_SET_insert( s, FK_SET_Elem_clone(e) );
  }
  //printf("s1 count: %d, s count: %d\n", s1->count, s->count);
  for (e = s2->head; e != NULL; e = e->next) {
    s = FK_SET_insert( s, FK_SET_Elem_clone(e) );
  }
  //printf("s2 count: %d, s count: %d\n", s2->count, s->count);
  return s;
}

static int FK_SET_Elem_isEmpty(SymbolNode * e) {
  if (e == NULL || e->snode == NULL || 
      strcmp(e->snode->symbol, "") == 0) return 1;
  return 0;
}

//
// Concatenate the SymbolList of e2 to e1.
// i.e., create a clone of e2's symbollist and concatenate to e1's symbollist.
// Tricky part: 
//   If a symbol node contains empty string, don't concat the node.
//   Note that the only ways an elem's string can grow are:
//   1) initialization from a single terminal, 2) concatenation,
//   So when empty elem happends, it must be initialized from empty string,
//   and contains only one symbol - the empty string.
//
static FK_SET_Elem * FK_SET_Elem_concat(FK_SET_Elem * e1, FK_SET_Elem * e2) {
  SymbolNode * s;
  if (e1->elem == NULL || FK_SET_Elem_isEmpty(e1->elem)) {
    e1->elem = cloneSymbolList(e2->elem);
    e1->length = e2->length;
  } else if ( e2->elem == NULL || FK_SET_Elem_isEmpty(e2->elem) ) {
    // do nothing.
  } else {
    for (s = e1->elem; s->next != NULL; s = s->next) {}
    s->next = cloneSymbolList(e2->elem);
    e1->length += e2->length;
  }

  return e1;
}

// truncate e such that it's elem length <= k.
static FK_SET_Elem * FK_SET_Elem_truncate(FK_SET_Elem * e, int k) {
  SymbolNode * s;
  int i;
  if (e == NULL || e->length <= k || k < 0) return e;

  if (k == 0) {
    freeSymbolNode(e->elem);
    e->elem = NULL;
  } else {
    s = e->elem;
    for (i = 0; i < k - 1; i ++) { s = s->next; }
    freeSymbolNodeList(s->next);
    s->next = NULL;
  }

  e->length = k;
  return e;
}


//
// Return 1 if e exists in s, 0 otherwise.
//
static int FK_SET_exist(FK_SET * s, FK_SET_Elem * e) {
  FK_SET_Elem * i;
  if (s == NULL || s->head == NULL || e == NULL) { puts("????????????");
    return 0;
  }

  for (i = s->head; i != NULL; i = i->next) {
    // "e != i" guarantees that e is not compared to itself.
    if (e != i && FK_SET_Elem_compare(e, i) == 0) return 1;
  }

  return 0;
}


//
// for each element of s, truncate by k.
// tricky part: if truncate causes 2 elements become the same, need to remove.
// the removal can be facilitated by the ordered property of the set.
//
static FK_SET * FK_SET_truncate(FK_SET * s, int k) {
  FK_SET_Elem * e, * e_prev;
  if (s == NULL || s->count == 0 || k < 0) return s;

  e_prev = NULL;
  e = s->head;

  while (e != NULL) {
    FK_SET_Elem_truncate(e, k);
    //if (FK_SET_Elem_compare(e_prev, e) == 0) {
    if ( FK_SET_exist(s, e) == 1 ) {
      // note s->count > 0, so e != NULL, thus e_prev != NULL.
      e_prev->next = e->next;
      FK_SET_Elem_destroy(& e);
      e = e_prev->next;
      s->count --;
    } else {
      e_prev = e;
      e = e->next;
    }
  }

  return s;
}
  

// test FK_SET creation and insert, FK_SET_Elem creation.
static void FK_SET_test1() {
  int i;
  FK_SET_Elem * e;
  FK_SET * s1;
  printf("==TEST FK_SET 1==\n");

  s1 = FK_SET_create();
  printf("\n--SET s1: \n"); FK_SET_dump(s1);

  for (i = 0; i < ParsingTblCols; i ++) {
    e = FK_SET_Elem_create(createSymbolNode(ParsingTblColHdr[i]));
    FK_SET_Elem_dump(e);
    FK_SET_insert(s1, e);
  }

  printf("\ninsert again\n");
  for (i = 0; i < ParsingTblCols; i ++) {
    e = FK_SET_Elem_create(createSymbolNode(ParsingTblColHdr[i]));
    FK_SET_Elem_dump(e);
    FK_SET_insert(s1, e);
  }

  printf("\n--SET s1: \n"); FK_SET_dump(s1);
  printf("\n==TEST FK_SET END==\n");
}

// test FK_SET union.
static void FK_SET_test2() {
  int i;
  FK_SET_Elem * e;
  FK_SET * s1, * s2, * s3;
  printf("==TEST FK_SET 2==\n");

  s1 = FK_SET_create();
  s2 = FK_SET_create();

  for (i = 0; i < 5; i ++) {
    e = FK_SET_Elem_create(createSymbolNode(ParsingTblColHdr[i]));
    FK_SET_Elem_dump(e);
    FK_SET_insert(s1, e);
  }

  for (i = 5; i < ParsingTblCols; i ++) {
    e = FK_SET_Elem_create(createSymbolNode(ParsingTblColHdr[i]));
    FK_SET_Elem_dump(e);
    FK_SET_insert(s2, e);
  }

  s3 = FK_SET_clone(s1);
  s3 = FK_SET_union(s3, s2);

  printf("\n--SET s1: \n"); FK_SET_dump(s1);
  printf("\n--SET s2: \n"); FK_SET_dump(s2);
  printf("\n--SET s3: \n"); FK_SET_dump(s3);

  printf("\n==TEST FK_SET END==\n");
}

// test FK_SET_Elem_concat.
static void FK_SET_test3() {
  int i;
  FK_SET_Elem * e, * f;
  FK_SET * s1;
  printf("==TEST FK_SET 3==\n");

  s1 = FK_SET_create();
  printf("\n--SET s1: \n"); FK_SET_dump(s1);

  e = FK_SET_Elem_create(createSymbolNode(ParsingTblColHdr[0]));

  for (i = 1; i < ParsingTblCols; i ++) {
    f = FK_SET_Elem_create(createSymbolNode(ParsingTblColHdr[i]));
    //printf("\n:"); FK_SET_Elem_dump(f);
    FK_SET_Elem_concat(e, f);
    //printf("\n::"); FK_SET_Elem_dump(e);
  }
  //puts("");

  FK_SET_Elem_truncate(e, 10); 
  printf("===after truncate, e.len = %d\n", e->length); FK_SET_Elem_dump(e);

  FK_SET_insert(s1, e);
  FK_SET_dump(s1);

  for (i = 0; i < 5; i ++) {
    e = FK_SET_Elem_create(createSymbolNode(ParsingTblColHdr[i]));
    FK_SET_Elem_dump(e);
    FK_SET_insert(s1, e);
  }
  FK_SET_dump(s1);

  FK_SET_truncate(s1, 0);
  FK_SET_dump(s1);

  printf("\n==TEST FK_SET END==\n");
}

// test clone.
static void FK_SET_test4() {
  int i;
  FK_SET_Elem * e;
  FK_SET * s1, * s2;
  printf("==TEST FK_SET 4==\n");

  s1 = FK_SET_create();
  printf("\n--SET s1: \n"); FK_SET_dump(s1);

  for (i = 0; i < ParsingTblCols; i ++) {
    e = FK_SET_Elem_create(createSymbolNode(ParsingTblColHdr[i]));
    FK_SET_Elem_dump(e);
    FK_SET_insert(s1, e);
  }

  s2 = FK_SET_clone(s1);

  printf("\n--SET s1: \n"); FK_SET_dump(s1);
  printf("\n--cloned SET s2: \n"); FK_SET_dump(s2);
 
  FK_SET_destroy(& s1); 
  printf("\n--SET s1: \n"); FK_SET_dump(s1);
  printf("\n--cloned SET s2: \n"); FK_SET_dump(s2);

  printf("\n==TEST FK_SET END==\n");
}

static void test_FK_SET() {
  //FK_SET_test1();
  //FK_SET_test2();
  //FK_SET_test3();
  //FK_SET_test4();
  //hashTbl_dump();
}

////////////////////////////////////
// Set for FIRST_K. END.
////////////////////////////////////


///////////////////////
// plus_k(). START.
///////////////////////

// do plus_k operation, and combine result into a new FK_SET.
static FK_SET * plus_k(FK_SET * s1, FK_SET * s2, int k) {
  FK_SET_Elem * i, * j, * e;
  FK_SET * s3 = FK_SET_create(); 

  for (i = s1->head; i != NULL; i = i->next) {
    for (j = s2->head; j != NULL; j = j->next) {
      e = FK_SET_Elem_clone(i);
      e = FK_SET_Elem_concat(e, j);
      FK_SET_insert(s3, e);
    }
  }

  FK_SET_truncate(s3, k);
  return s3;
}

static void test_plus_k() {
  FK_SET * s1, * s2, * s3, * s4, * ss;
  FK_SET_Elem * e, * f;
  int i;

#if DEBUG & 2
  printf("enter test_plus_k()\n");
#endif

  s1 = FK_SET_create();
  f = FK_SET_Elem_create(createSymbolNode(hashTbl_insert("")));
  FK_SET_insert(s1, f);
  for (i = 1; i < 4; i ++) {
    e = FK_SET_Elem_create(createSymbolNode(ParsingTblColHdr[i]));
    FK_SET_insert(s1, e);
  }

  s2 = FK_SET_clone(s1);
  s3 = FK_SET_clone(s1);

  s4 = FK_SET_create();
  FK_SET_insert(s4, FK_SET_Elem_create(createSymbolNode(ParsingTblColHdr[4])));

  ss = plus_k( plus_k( plus_k(s1, s2, 2), s3, 2), s4, 2);

  FK_SET_dump(s1);
  FK_SET_dump(s2);
  FK_SET_dump(s3);
  FK_SET_dump(s4);
  FK_SET_dump(ss);
}

///////////////////////
// plus_k(). END.
///////////////////////

// F_Table for terminals. Only 1 row.
// Empty string is also included as the first element of this array.
static FK_SET ** F_T;

// A row for the F_Table of non-terminals. rows 1, 2, ... until converge.
struct NT_F_TblRow {
  FK_SET ** row;
  int row_num; 
  struct NT_F_TblRow * next;
};
typedef struct NT_F_TblRow F_NT_Row;

// F_Table for non-terminals as a linked list..
struct {
  F_NT_Row * head;
  //F_NT_Row * tail;
  int count;
} F_NT_List;

// F_Table for non-terminals as an array, for easy random access.
static F_NT_Row ** F_NT;
static int F_NT_size; // max size of the F_NT array.

static F_NT_Row * F_NT_Row_create(int row_num) {
  int i;
  F_NT_Row * r;
  HYY_NEW(r, F_NT_Row, 1);
  HYY_NEW(r->row, FK_SET *, grammar.non_terminal_count);
  for (i = 0; i < grammar.non_terminal_count; i ++) {
    r->row[i] = FK_SET_create();
  }
  r->row_num = row_num;
  r->next = NULL;
  return r;
}

// return 1 if true, 0 if false.
static int F_NT_Row_equal(F_NT_Row * r1, F_NT_Row * r2) {
  int i, len = grammar.non_terminal_count;
  FK_SET * s1, * s2;
  for (i = 0; i < len; i ++) {
    s1 = r1->row[i];
    s2 = r2->row[i];
    if (FK_SET_compare(s1, s2) != 0) return 0;
  }
  return 1;
}

static void F_T_dump() {
  int i;
  printf("==F_T==:\n"); 

  // header.
  printf("    ");
  printf(" \"\" | "); // for empty string.
  for (i = 0; i < grammar.terminal_count; i ++) {
    printf( " %s  | ", ParsingTblColHdr[i+1]->symbol );
  }
  puts("\n----------------------------------------------------");

  printf("F0: ");
  for (i = 0; i <= grammar.terminal_count; i ++) {
    FK_SET_dump2(F_T[i]); printf(" | ");
  }
  puts("");
}

static void F_NT_Row_dump(F_NT_Row * r) {
  int i;
  for (i = 0; i < grammar.non_terminal_count -1; i ++) {
    FK_SET_dump2(r->row[i]); printf(" | ");
  }
  puts("");
}

static void F_NT_List_dump() {
  int i, ct = 0;
  F_NT_Row * r;
  printf("==F_NT_List==:\n");

  // header.
  printf("    ");
  for (i = 0; i < grammar.non_terminal_count - 1; i ++) {
    printf( " %s  | ", 
            ParsingTblColHdr[grammar.terminal_count + 2 + i]->symbol );
  }
  puts("\n----------------------------------------------------");

  for (r = F_NT_List.head; r != NULL; r = r->next) {
    printf("F%d: ", ct ++);
    F_NT_Row_dump(r);
  }
  puts("");
}

// Given the RHS of a production, return x-thead of it, x <= k.
static SymbolList getRHS_KTHead(SymbolList RHS, int k) {
  // add string of up to K starting terminals to F0.
  SymbolNode * s = cloneSymbolList(RHS);
  SymbolNode * t, * t_prev;
  int i;
  for (t_prev = s, t = s->next, i = 1; 
       i < k && t != NULL && isTerminal(t->snode); 
       t_prev = t, t = t->next, i ++) {}
  freeSymbolNodeList(t);
  t_prev->next = NULL;

  return s;
}

//
// Build F0 for non-terminal.
// Parameters:
//   s - the FK_SET for F0 of this NT symbol (n).
//
static void build_F_NT_List_F0(FK_SET * s, SymbolTblNode * n, int K) {
  Production * p;
  SymbolList pR; // right hand side of p.
  RuleIDNode * r = n->ruleIDList;

  if (K == 0) return;

#if DEBUG_FK & 4
  printf("\n---NT(%s): ", n->symbol);
#endif

  for (; r != NULL; r = r->next) {
    p = grammar.rules[r->ruleID];
    pR = p->nRHS_head;
#if DEBUG_FK & 4
    printf("%d(rhs.cnt=%d).", r->ruleID, p->RHS_count);
#endif
    //writeSymbolList(pR, "rhs");
    if (p->RHS_count == 0) { // RHS is empty, i.e., empty production.
      // add empty string symbol node.
      FK_SET_insert(s, 
        FK_SET_Elem_create(createSymbolNode(hashTbl_insert(""))));
    }
    else if (isTerminal( p->nRHS_head->snode )) {
      SymbolNode * sn = getRHS_KTHead(p->nRHS_head, K);
      FK_SET_insert(s, FK_SET_Elem_create(sn));
    } else { // RHS first symbol is non-terminal.
      // do nothing.
    }
    
  }
  //puts("");
}

static FK_SET * get_Fi_1(SymbolTblNode * n, F_NT_Row * Fi_1) {
  if (isTerminal(n)) {
    return F_T[n->seq];
  } else { // is non-terminal
    return Fi_1->row[n->seq - grammar.terminal_count - 2];
  }
}

static void build_F_NT_List_Fi(
    FK_SET * Fi, F_NT_Row * Fi_1, int ni, SymbolTblNode * n, int K) {
  Production * p;
  SymbolList pR; // right hand side of p.
  RuleIDNode * r = n->ruleIDList;

  if (K == 0) return;

#if DEBUG_FK & 4
  printf("\n---NT(%s): ", n->symbol);
#endif

  for (; r != NULL; r = r->next) {
    p = grammar.rules[r->ruleID];
    pR = p->nRHS_head;
#if DEBUG_FK & 4
    printf("%d(rhs.cnt=%d).", r->ruleID, p->RHS_count);
#endif
    //writeSymbolList(pR, "rhs");
    if (p->RHS_count == 0) { // RHS is empty, i.e., empty production.
      // do nothing.
    } else if (p->RHS_count == 1) {
      Fi = FK_SET_union(Fi, get_Fi_1(p->nRHS_head->snode, Fi_1));
    } else { // RHS 
      FK_SET * s = get_Fi_1(p->nRHS_head->snode, Fi_1);
      SymbolNode * t;
      for (t = p->nRHS_head->next; t != NULL; t = t->next) {
        s = plus_k(s, get_Fi_1(t->snode, Fi_1), K);
      }
      Fi = FK_SET_union(Fi, s);
    }
  }
  //puts("");
}

static void build_F_NT_List(int K) {
  F_NT_Row * Fi; // new row F(i).
  F_NT_Row * Fi_1 = F_NT_List.head; // old row F(i-1).
  SymbolTblNode * n;
  int ct = 0, col, ni;
  int done = 0;

  while(! done) {
    ct ++; // build the i-th row.
    Fi = F_NT_Row_create(ct);

    for (col = 0; col < ParsingTblCols; col ++) {
      n = ParsingTblColHdr[col];
      if (isNonTerminal(n) && ! isGoalSymbol(n)) { 
        //printf("n:%s(%d), ", n->symbol, n->seq);
        ni = n->seq - grammar.terminal_count -2;
        Fi->row[ni] = FK_SET_clone(Fi_1->row[ni]);
        build_F_NT_List_Fi(Fi->row[ni], Fi_1, ni, n, K);
      }
    } // end of for.

    if (F_NT_Row_equal(Fi, Fi_1)) break; // done = 1;

    Fi_1->next = Fi;
    Fi_1 = Fi;
    F_NT_List.count ++;
  }

  F_NT_size = ct;

#if DEBUG_FK & 4
  printf("F_NT size: %d\n", F_NT_size);
#endif
}

static void build_F_NT_Table() {
  int i;
  F_NT_Row * r;
  HYY_NEW(F_NT, F_NT_Row *, F_NT_size);
  for (i = 0, r = F_NT_List.head; i < F_NT_size; i ++, r = r->next) {
    F_NT[i] = r;
  }
}

static void F_NT_dump() {
  int i;

  printf("==F_NT Table==:\n");

  // header.
  printf("    ");
  for (i = 0; i < grammar.non_terminal_count - 1; i ++) {
    printf( " %s  | ",
            ParsingTblColHdr[grammar.terminal_count + 2 + i]->symbol );
  }
  puts("\n----------------------------------------------------");

  for (i = 0; i < F_NT_size; i++) {
    printf("F%d: ", i);
    F_NT_Row_dump(F_NT[i]);
  }
  puts("");
}

static FK_SET * F_NT_get(SymbolTblNode * n, int k) {
  if (k < 0) return NULL;
  if (k >= F_NT_size) k = F_NT_size - 1; // return last entry.

  return F_NT[k]->row[n->seq - grammar.terminal_count - 2];
}

//
// Get Fk(X), X can be terminal or non-terminal.
//
static FK_SET * F_get(SymbolTblNode * n, int k) {
  // spacial case: empty string node.

  if (isTerminal(n)) {
    return F_T[n->seq];
  } else if (strcmp(n->symbol, "") == 0) {
    return F_T[0]; 
  } else { // is non-terminal
    // FIRST_k_{X} = Fi(X), i == F_NT_size - 1, is the i when converge.
    //return F_NT_get(n, k); // this is incorrect, say when k = 1.
    return F_NT_get(n, F_NT_size - 1);
  }
}

static void test_F_get() {
  int col, i;
  SymbolTblNode * n;

  for (col = 0; col < ParsingTblCols; col ++) {
    n = ParsingTblColHdr[col];
    if (isGoalSymbol(n)) { } // do nothing.
    else if (strcmp(strEnd, n->symbol) == 0) { } // do nothing.
    else {
      printf("symbol : %s(%d): \n", n->symbol, n->seq);
      for (i = 0; i <= F_NT_size; i ++) {
        FK_SET_dump( F_get(n, i) );
        puts("");
      }
      puts("---------------");
    }
  } // end of for.
  puts("");
}

static FK_SET * build_F_strEmpty() {
  FK_SET * s;
  FK_SET_Elem * e;

  HYY_NEW(e, FK_SET_Elem, 1);
  e->elem = createSymbolNode(hashTbl_insert(""));
  e->length = 1;
  e->next = NULL;

  HYY_NEW(s, FK_SET, 1);
  s->head = s->tail = e;
  s->count = 1;
  return s;
}

static void build_F_Table(int K) {
  int i, col, count = 0;
  int col_size = ParsingTblCols;
  SymbolTblNode * n;

#if DEBUG_FK & 2
  printf("enter build_F_Table()\n");
#endif

  // 1) create the FK_SET for empty string.
  build_F_strEmpty();

#if DEBUG_FK & 2
  printf("T.count = %d, NT.count = %d\n", 
         grammar.terminal_count, grammar.non_terminal_count);
#endif

  // 2) get terminal list. For a terminal with seq, entry is (seq - 1).
  HYY_NEW(F_T, FK_SET *, grammar.terminal_count + 1); // +1 for empty string.
  F_T[0] = build_F_strEmpty();
  for (i = 1; i <= grammar.terminal_count; i ++) {
    F_T[i] = FK_SET_create();
  }

  // get non-terminal list. For a non-terminal with seq, 
  // entry is (seq - grammar.terminal_count - 2). -2 for $end and $accept.
  // Initialize the first row of F_NT.
  F_NT_List.head = F_NT_Row_create(0);

  // 3) build F table for non-terminals.
  // Initialize the first row of F_T and F_NT_List.
  for (col = 0; col < ParsingTblCols; col ++) {
    n = ParsingTblColHdr[col];
    if (isGoalSymbol(n)) { 
      // printf("g:%s(%d), ", n->symbol, n->seq); 
    } // $accept
    else if (strcmp(strEnd, n->symbol) == 0) {
      //printf("e:%s(%d), ", n->symbol, n->seq); 
    } // $end
    else if (isTerminal(n)) { 
      // printf("t:%s(%d), ", n->symbol, n->seq); 
      FK_SET_insert(F_T[n->seq], FK_SET_Elem_create(createSymbolNode(n)));
    }
    else if (isNonTerminal(n)) { 
      // printf("n:%s(%d), ", n->symbol, n->seq); 
      build_F_NT_List_F0(
        F_NT_List.head->row[n->seq - grammar.terminal_count -2], n, K);
    }
    else { printf("%s (should NOT happen), ", n->symbol); }
  } // end of for.
  //puts("");

  //F_NT_List.tail = F_NT_List.head;
  F_NT_List.count = 1;

  // build Fi for non-terminals, i >= 1.
  build_F_NT_List(K);
  build_F_NT_Table();

#if DEBUG_FK & 2
  F_T_dump();
  F_NT_List_dump();
  F_NT_dump();
  //test_F_get();
#endif
}


////////////////////////////////////////////////
// Public interface functions.
////////////////////////////////////////////////

static int F_Table_Built = 0; // Do lazy initialization of F_Table.

//
// The API function of Aho & Ullman's First_k_{alpha} algorithm.
//
FK_SET * FirstK(SymbolList alpha, int k) {
  SymbolNode * n;
  FK_SET * s;

  if (! F_Table_Built) { 
    build_F_Table(5);
    F_Table_Built = 1;
  }

  n = alpha;
  s = F_get(n->snode, k);
  for (n = n->next; n != NULL; n = n->next) {
    s = plus_k(s, F_get(n->snode, k), k);
  }

#if DEBUG_FK
  writeSymbolList(alpha, "alpha");
  printf("FIRST(alpha, %d): [%d-thead: %d]\n", k, k, FK_SET_count_len(s, k)); 
  FK_SET_dump(s);
#endif

  return s;
}

// return number of elements whose length is k.
int FK_SET_count_len(FK_SET * s, int len) {
  FK_SET_Elem * e;
  int k = 0;
  if (s == NULL) return 0;

  e = s->head;
  if (FK_SET_Elem_isEmpty(e->elem)) { e = e->next; } // don't count empty string as thead.

  for (; e != NULL; e = e->next) {
    if (e->length == len) k ++;
  }
  return k;
}

//
// Show content of a FK_SET.
//
void FK_SET_dump(FK_SET * s) {
  FK_SET_Elem * e;
  int i = 0;
  if (s == NULL) return;

  e = s->head;
  printf("{");
  for (; e != NULL; e = e->next) {
    printf("\n  %2d:", ++i);
    FK_SET_Elem_dump(e);
  }
  printf("\n} (set elem count: %d)\n", s->count);
}

//
// Show the Fi(X) table for all symbols.
//
void F_dump() {
  F_T_dump();
  F_NT_dump();
}


void test_firstk(SymbolList alpha) {
  int i;
  printf("\nenter test_firstk()\n");

  //test_FK_SET();
  //test_plus_k();

  for (i = 0; i <= 5; i ++) { 
    FirstK(alpha, i);
  }

  //F_dump();
}
